import copy
from ..utils.config import load_all_configs
from .route import route_vac
from ..custom_logging import log
from ruamel.yaml import YAML
from io import StringIO
# check it here:
# https://editor.swagger.io/
from functools import lru_cache
import uuid

try:
    from google.cloud import servicecontrol_v1
    from google.protobuf.timestamp_pb2 import Timestamp
except ImportError:
    servicecontrol_v1 = None

def validate_api_key(api_key: str, service_name: str) -> bool:
    """
    Validate an API key against the service name e.g. 'endpoints-xxxx.a.run.app'
    """
    if not servicecontrol_v1:
        raise ImportError("Cloud Endpoints API key validation is required, install via `pip install sunholo[gcp]`")

    return _validate_api_key_cached(api_key, service_name)


@lru_cache(maxsize=1024)
def _validate_api_key_cached(api_key: str, service_name: str) -> bool:
    client = servicecontrol_v1.ServiceControllerClient()
    # Create a timestamp for the current time
    start_time = Timestamp()
    start_time.GetCurrentTime()

    request = servicecontrol_v1.CheckRequest(
        service_name=service_name,
        operation=servicecontrol_v1.Operation(
            operation_name=service_name.upper(),
            operation_id=str(uuid.uuid4()),
            consumer_id=f'api_key:{api_key}',
            start_time=start_time,
        )
    )
    log.debug(f"API_KEY checkRequest {request}")

    response = client.check(request=request)

    # Check if there are any check errors
    if response.check_errors:
        log.debug(f"API_KEY validation failed with errors: {response.check_errors}")
        return False

    log.debug(f"API_KEY validation succeeded - {response}")
    return True

def config_to_swagger():
    """
    Load configuration files and generate a Swagger specification.

    This function loads the 'vacConfig' and 'agentConfig' configuration files,
    validates their presence, and then generates a Swagger specification
    based on these configurations.

    Returns:
        str: The generated Swagger specification in YAML format.

    Raises:
        ValueError: If 'vacConfig' or 'agentConfig' is not loaded.

    Example:
    ```python
        swagger_yaml = config_to_swagger()
        print(swagger_yaml)
    ```
    """
    configs = load_all_configs()

    vac_config = configs.get('vacConfig')
    agent_config = configs.get('agentConfig')

    if not vac_config:
        raise ValueError("Need valid 'vacConfig' loaded")
    
    if not agent_config:
        raise ValueError("Need valid 'agentConfig' loaded")
    
    swag = generate_swagger(vac_config, agent_config)

    return swag

def generate_swagger(vac_config, agent_config):
    """
    Generate a Swagger specification based on the provided configurations.

    This function creates a Swagger specification using the provided 'vacConfig'
    and 'agentConfig'. It dynamically builds paths and responses based on the
    configurations.

    Args:
        vac_config (dict): The VAC configuration.
        agent_config (dict): The agent configuration.

    Returns:
        str: The generated Swagger specification in YAML format.

    Example:
    ```python
        vac_config = {
            'vac': {
                'service1': {
                    'llm': 'vertex',
                    'model': 'gemini-1.5-flash-001',
                    'agent': 'langserve'
                },
                'service2': {
                    'llm': 'openai',
                    'agent': 'crewai',
                    'secrets': ['OPENAI_API_KEY']
                }
            }
        }
        
        agent_config = {
            'agents': {
                'default': {
                    'stream': "{stem}/vac/streaming/{vector_name}",
                    'invoke': "{stem}/vac/{vector_name}",
                    'post': {
                        'stream': "{stem}/vac/streaming/{vector_name}",
                        'invoke': "{stem}/vac/{vector_name}",
                        'openai': "{stem}/openai/v1/chat/completions",
                        'openai-vac': "{stem}/openai/v1/chat/completions/{vector_name}"
                    },
                    'get': {
                        'home': "{stem}/",
                        'health': "{stem}/health"
                    },
                    'response': {
                        'invoke': {
                            '200': {
                                'description': 'Successful invocation response',
                                'schema': {
                                    'type': 'object',
                                    'properties': {
                                        'answer': {'type': 'string'},
                                        'source_documents': {
                                            'type': 'array',
                                            'items': {
                                                'type': 'object',
                                                'properties': {
                                                    'page_content': {'type': 'string'},
                                                    'metadata': {'type': 'string'}
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        
        swagger_yaml = generate_swagger(vac_config, agent_config)
        print(swagger_yaml)
        ```
    """
    swagger_template = {
        'swagger': '2.0',
        'info': {
            'title': 'Sunholo Multivac API - ${_BRANCH_NAME}',
            'description': 'Multivac (Multi-VAC) Cloud. Generate API keys (MULTIVAC_API_KEY) to enable access to private services. See more at https://dev.sunholo.com/',
            'version': '0.1.0'
        },
        'host': '${_ENDPOINTS_HOST}',
        'basePath': '/v1',
        'schemes': ['https'],
        'produces': ['application/json'],
        'paths': {},
        'securityDefinitions': { 
            'ApiKeyAuth': {        # For private VAC endpoint
                'type': 'apiKey',
                'name': 'x-api-key', # Custom header name for API key
                'in': 'header'
            },
            'None': {              # For public documentation
                'type': 'apiKey',
                'name': 'allow',    # Dummy parameter for public access
                'in': 'query'       # Use query parameter to avoid interfering with other headers
            }
        }
    }
    vac_services = vac_config['vac']
    
    def configure_agent(vector_name, config, agent_config_paths):
        agent_type = config['agent']
        log.info(f'Configuring swagger for agent_type: {agent_type} for vector_name: {vector_name}')
        try:
            stem = route_vac(vector_name).strip()
        except ValueError:
            log.warning(f"Failed to find URL stem for {vector_name}/{agent_type} - skipping")
            return

        for method, endpoints in agent_config_paths.items():
            do_auth = True

            if method not in ['get', 'post', 'get-auth', 'post-noauth']:
                continue

            if method == 'get-auth':
                do_auth = True
                method = 'get'
            elif method == 'post-noauth':
                do_auth = False
                method = 'post'
            elif method == 'get':
                do_auth = False
            elif method == 'post':
                do_auth = True

            for endpoint_key, endpoint_template in endpoints.items():
                endpoint_template = endpoint_template.strip()
                endpoint_address = endpoint_template.replace("{stem}", stem).replace("{vector_name}", vector_name).strip()
                endpoint_path = endpoint_template.replace("{stem}", f"/{agent_type}").replace("{vector_name}", vector_name).strip()
                log.debug(f"Endpoint_template: {endpoint_template}")
                log.debug(f"endpoint address: {endpoint_address}")
                log.debug(f"endpoint_path: {endpoint_path}")
                if endpoint_path not in swagger_template['paths']:
                    swagger_template['paths'][endpoint_path] = {}

                operation_id = f"{method}_{agent_type}_{endpoint_key}_{vector_name}"

                security = [{'ApiKeyAuth': []}] if do_auth else [{'None': []}]

                swagger_template['paths'][endpoint_path][method] = {
                    'summary': f"{method.capitalize()} {vector_name}",
                    'operationId': operation_id,
                    'x-google-backend': {
                        'address': endpoint_address,
                        'protocol': 'h2',
                        'deadline':  180000  # Timeout of 3 minutes (1,800,00 milliseconds)
                    },
                    'security': security,
                    'responses': copy.deepcopy(agent_config_paths.get('response', {}).get(endpoint_key, {
                        '200': {
                            'description': 'Default - A successful response',
                            'schema': {
                                'type': 'string'
                            }
                        }
                    }))
                }

    for vector_name, config in vac_services.items():
        agent_type = config['agent']
        agent_config_paths = agent_config['agents'].get(agent_type, {})
        configure_agent(vector_name, config, agent_config_paths)

    default_agent_config = agent_config['agents'].get('default', {})

    for vector_name, config in vac_services.items():
        agent_type = config['agent']
        if agent_type in agent_config['agents']:
            continue
        log.info(f'Applying default configuration for agent_type: {agent_type} for vector_name: {vector_name}')
        configure_agent(vector_name, config, default_agent_config)

    yaml = YAML()
    yaml.width = 4096

    string_stream = StringIO()
    yaml.dump(swagger_template, string_stream)
    yaml_string = string_stream.getvalue()

    try:
        from openapi_spec_validator import validate
        swagger_dict = yaml.load(yaml_string)
        swagger_dict['host'] = 'api.sunholo.com'
        validated = validate(swagger_dict)
        log.info(f'swagger validation errors: {validated}')
    except ImportError:
        log.error("Must have openapi-spec-validator installed to validate - install via `pip install sunholo'[tools]'`")

    return yaml_string


from ..custom_logging import log
from ..gcs import get_bytes_from_gcs

from functools import partial
import mimetypes
import asyncio
import tempfile
import re
import os
import traceback
try:
    import google.generativeai as genai
except ImportError:
    genai = None

DOCUMENT_MIMES = [
    'application/pdf',
    'application/x-javascript',
    'text/javascript',
    'application/x-python',
    'text/x-python',
    'text/plain',
    'text/html',
    'text/css',
    'text/md',
    'text/csv',
    'text/xml',
    'text/rtf'
    ]

IMAGE_MIMES = [
    'image/png',
    'image/jpeg',
    'image/webp',
    'image/heic',
    'image/heif',
]

VIDEO_MIMES = [
    'video/mp4',
    'video/mpeg',
    'video/mov',
    'video/avi',
    'video/x-flv',
    'video/mpg',
    'video/webm',
    'video/wmv',
    'video/3gpp'
]

AUDIO_MIMES = [
    'audio/wav',
    'audio/mp3',
    'audio/mpeg', #added
    'audio/aiff',
    'audio/aac',
    'audio/ogg',
    'audio/flac',    
]

ALLOWED_MIME_TYPES = set(AUDIO_MIMES + VIDEO_MIMES + IMAGE_MIMES + DOCUMENT_MIMES)

# 'documents': 
# [
#   {'storagePath': 'users/UQcKi4u7s...dsd.png', 
#    'url': 'https://firebasestorage.googleapis.com/v0/b/multi...', 
#    'contentType': 'image/png', 
#    'type': 'image', 
#    'name': 'multivac-data-architecture.png'}, 
#   {'storagePath': 'users/UQc...3dc59e1.jpg', 
#   'type': 'image', 
#   'name': 'holosun-circle.jpg', 
#   'url': 'https://firebasestorage.googleapis.com/v0/b/multiv...', 
#   'contentType': 'image/jpeg'}
# ]

def sanitize_file(filename):
    # Split the filename into name and extension
    name, extension = os.path.splitext(filename)
    
    # Sanitize the name by removing invalid characters and converting to lowercase
    sanitized_name = re.sub(r'[^a-z0-9-]', '', name.lower())
    sanitized_name = re.sub(r'^-+|-+$', '', sanitized_name)  # Remove leading or trailing dashes
    
    # Reattach the original extension
    return sanitized_name[:40]

async def construct_file_content(gs_list, bucket:str):
    """
    Args:
    - gs_list: a list of dicts representing files in a bucket
       - contentType: The content type of the file on GCS
       - storagePath: The path in the bucket
       - name: The name of the file
    - bucket: The bucket the files are in

    """
    
    file_list = []
    for element in gs_list:

        the_mime_type = element.get('contentType')
        if the_mime_type is None:
            continue
        if element.get('storagePath') is None:
            continue
        if the_mime_type in ALLOWED_MIME_TYPES:
            file_list.append(element)
        else:
            log.warning(f'{the_mime_type} is not in allowed MIME types for {element.get("name")}')
    
    if not file_list:
        return {"role": "user", "parts": [{"text": "No eligible contentTypes were found"}]}

    content = []
    
    # Loop through the valid files and process them
    tasks = []
    for file_info in file_list:
        img_url = f"gs://{bucket}/{file_info['storagePath']}"
        display_url = file_info.get('url')
        mime_type = file_info['contentType']
        name = sanitize_file(file_info['name'])
        display_name = file_info['name']
        log.info(f"Processing {name=} {display_name=}")
        try:
            myfile = genai.get_file(name)
            content.append(
                {"role": "user", "parts": [
                    {"file_data": myfile}, 
                    {"text": f"You have been given the ability to work with file {display_name=} with {mime_type=} {display_url=}"}
                    ]
                })
            log.info(f"Found existing genai.get_file {name=}")
            
        except Exception as e:
            log.info(f"Not found checking genai.get_file: '{name}' {str(e)}")
            tasks.append(
                download_gcs_upload_genai(img_url, 
                                          mime_type=mime_type, 
                                          name=name, 
                                          display_url=display_url, 
                                          display_name=display_name)
                                          )

    # Run all tasks in parallel
    if tasks:
        task_content = await asyncio.gather(*tasks)
        content.extend(task_content)

    return content

# Helper function to handle each file download with error handling
async def download_file_with_error_handling(img_url, mime_type, name):
    try:
        return await download_gcs_upload_genai(img_url, mime_type, name)
    except Exception as err:
        msg= f"Error processing file from {img_url}: {str(err)}"
        log.error(msg)
        return {"role": "user", "parts": [{"text": msg}]}

async def download_gcs_upload_genai(img_url, 
                                    mime_type, 
                                    name=None, 
                                    display_url=None, 
                                    display_name=None, 
                                    retries=3, delay=2):
    import aiofiles
    from google.generativeai.types import file_types
    """
    Downloads and uploads a file with retries in case of failure.
    
    Args:
      - img_url: str The URL of the file to download.
      - mime_type: str The MIME type of the file.
      - name: str Optional name, else a random one will be created
      - retries: int Number of retry attempts before failing.
      - delay: int Initial delay between retries, exponentially increasing.
      
    Returns:
      - downloaded_content: The result of the file upload if successful.
    """
    for attempt in range(retries):
        try:
            log.info(f"Upload {attempt} for {img_url=}")
            # Download the file bytes asynchronously
            file_bytes = await asyncio.to_thread(get_bytes_from_gcs, img_url)
            if not file_bytes:
                msg = f"Failed to download file from {img_url}: got None"
                log.warning(msg)
                return {"role": "user", "parts": [{"text": msg}]}
            
            # Log the size of the file bytes
            file_size = len(file_bytes)
            log.info(f"Downloaded file size for {img_url}: {file_size} bytes")

            if file_size > 19434343:
                log.warning(f"File size for {img_url}: {file_size} is too big.")
                msg =  f"The file for {img_url} is too large ({file_size} bytes) to be used directly.  Use RAG instead or {display_url=}"
                return {"role": "user", "parts": [{"text": msg}]}
            
            extension = mimetypes.guess_extension(mime_type)
            
            # Use aiofiles for asynchronous file operations
            temp_file = tempfile.NamedTemporaryFile(delete=False, suffix=extension)
            downloaded_file = temp_file.name

            sanitized_file = sanitize_file(downloaded_file)

            log.info(f"Writing file {sanitized_file}")
            async with aiofiles.open(sanitized_file, 'wb') as f:
                await f.write(file_bytes)

            # Upload the file and get its content reference
            try:
                downloaded_content: file_types.File = await asyncio.to_thread(
                    partial(genai.upload_file, name=name, mime_type=mime_type, display_name=display_name), 
                    sanitized_file
                    )
                return {"role": "user", "parts": [{"file_data": downloaded_content}, 
                                                  {"text": f"You have been given the ability to read and work with filename '{display_name=}' with {mime_type=} {display_url=}"}
                                                  ]}
            except Exception as err:
                msg = f"Could not upload {sanitized_file} to genai.upload_file: {str(err)} {traceback.format_exc()} {display_url=}"
                log.error(msg)
                return {"role": "user", "parts": [{"text": msg}]}
        
        except Exception as err:
            log.error(f"Error processing file {img_url} {mime_type=} on attempt {attempt + 1}/{retries}: {str(err)}")

            if attempt < retries - 1:
                log.info(f"Retrying in {delay} seconds...")
                await asyncio.sleep(delay)
                delay *= 2  # Exponential backoff
            else:
                raise err  # Raise the error after max retries




import subprocess
import os
import signal
import json
import socket

from .sun_rich import console
from rich.table import Table
from rich import print

PROXY_TRACKER_FILE = os.path.join(os.path.expanduser("~"), '.sunholo_proxy_tracker.json')
DEFAULT_PORT = 8080


def is_port_in_use(port):
    """Check if a given port is in use on the localhost."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0

def get_next_available_port(proxies, default_port):
    """
    Get the next available port starting from the default port.

    Args:
        proxies (dict): Current proxies with their assigned ports.
        default_port (int): Default starting port.

    Returns:
        int: The next available port.
    """
    used_ports = {info["port"] for info in proxies.values()}
    port = default_port
    while port in used_ports or is_port_in_use(port):
        port += 1
    return port

def check_gcloud():
    """
    Checks if gcloud is installed and authenticated.

    Returns:
        bool: True if gcloud is installed and authenticated, False otherwise.
    """
    try:
        # Check if gcloud is installed
        result = subprocess.run(["gcloud", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=30)
        if result.returncode != 0:
            print("[bold red]ERROR: gcloud is not installed or not found in PATH.[/bold red]")
            return False

        # Check if gcloud is authenticated
        result = subprocess.run(["gcloud", "auth", "list"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=30)
        if result.returncode != 0 or "ACTIVE" not in result.stdout.decode():
            print("ERROR: gcloud is not authenticated. Please run 'gcloud auth login'.")
            return False

        print("gcloud is installed and authenticated.")
        return True
    except Exception as e:
        print(f"ERROR: An unexpected error occurred: {e}")
        return False

def is_process_running(pid):
    try:
        os.kill(pid, 0)
        return True
    except OSError:
        print("WARNING: VAC Proxy lost connection")
        return False

def load_proxies():
    if os.path.exists(PROXY_TRACKER_FILE):
        with open(PROXY_TRACKER_FILE, 'r') as file:
            return json.load(file)
    return {}

def clean_proxy_list():
    proxies = load_proxies()
    updated_proxies = {k: v for k, v in proxies.items() if is_process_running(v["pid"])}
    if len(proxies) != len(updated_proxies):
        save_proxies(updated_proxies)
        
    return updated_proxies

def save_proxies(proxies):
    with open(PROXY_TRACKER_FILE, 'w') as file:
        json.dump(proxies, file, indent=4)

def start_proxy(service_name, region, project, port=None, local=False, app_type=None, app_folder=None, log_file=None):
    """
    Starts the gcloud proxy to the Cloud Run service and stores the PID.

    Args:
        service_name (str): Name of the Cloud Run service.
        region (str): Region of the Cloud Run service.
        project (str): GCP project of the Cloud Run service.
        port (int, optional): Port to run the proxy on. If not provided, auto-assigns the next available port.
    """
    proxies = clean_proxy_list()

    if service_name in proxies:
        console.print(f"Proxy for service [bold orange]'{service_name}'[/bold orange] is already running on port {proxies[service_name]['port']}.")
        return

    if not port:
        port = get_next_available_port(proxies, DEFAULT_PORT)
    
    if local:
        start_local(service_name, port, app_type, app_folder, log_file)
    else:
        command = [
            "gcloud", "run", "services", "proxy", service_name,
            "--region", region,
            "--project", project,
            "--port", str(port)
        ]
        if log_file:
            log_file_path = os.path.join(app_folder, f"{service_name}_log.txt")
            with open(log_file_path, 'a') as logf:
                process = subprocess.Popen(command, stdout=logf, stderr=logf, preexec_fn=os.setpgrp)
        else:
            log_file_path = "No log file specified"
            with open(os.devnull, 'w') as devnull:
                process = subprocess.Popen(command, stdout=devnull, stderr=devnull, preexec_fn=os.setpgrp)
            
        proxies[service_name] = {
            "pid": process.pid,
            "port": port,
            "local": "No",
            "logs": log_file_path
        }
        save_proxies(proxies)
        
        console.print(f"Proxy for [bold orange]'{service_name}'[/bold orange] setup complete on port {port}")
        list_proxies()

    return f"http://127.0.0.1:{port}"


def stop_proxy(service_name, stop_local=True):
    """
    Stops the gcloud proxy to the Cloud Run service using the stored PID.

    Args:
        service_name (str): Name of the Cloud Run service.
        stop_local (bool): Whether to stop locally running services or not. Defaults to True.
    """
    proxies = clean_proxy_list()

    # ps aux | grep gcloud
    # kill PID

    if service_name not in proxies:
        return

    if not stop_local:
        local = proxies[service_name]["local"]
        if local != "No":
            console.print(f"Not stopping local VAC running on: {local}")
    else:
        pid = proxies[service_name]["pid"]
        try:
            os.kill(pid, signal.SIGTERM)
            del proxies[service_name]
            save_proxies(proxies)
            console.print(f"Proxy for [bold orange]'{service_name}'[bold orange] stopped.")
        except ProcessLookupError:
            console.print(f"No process found with PID: {pid}")
        except Exception as e:
            console.print(f"[bold red]Error stopping proxy for {service_name}: {e}[/bold red]")
    
    list_proxies()

def stop_all_proxies():
    """
    Stops all running gcloud proxies.
    """
    proxies = clean_proxy_list()

    for service_name, info in proxies.items():
        pid = info["pid"]
        try:
            os.kill(pid, signal.SIGTERM)
            print(f"Proxy for [bold orange]'{service_name}'[/bold orange] stopped.")
        except ProcessLookupError:
            print(f"No process found with PID: {pid}")
        except Exception as e:
            print(f"Error stopping proxy for [bold orange]'{service_name}'[/bold orange]: {e}")
    
    save_proxies({})

    list_proxies()

def list_proxies():
    """
    Lists all running proxies.
    """
    with console.status("[bold orange]Listing Proxies[/bold orange]", spinner="star"):
        proxies = clean_proxy_list()

    if not proxies:
        print("No proxies currently running.")
    else:
        table = Table(title="VAC Proxies - `sunholo proxy list`")
        table.add_column("VAC")
        table.add_column("Port")
        table.add_column("PID")
        table.add_column("URL")
        table.add_column("Local")
        table.add_column("Logs")
        
        for service_name, info in proxies.items():
            url = f"http://127.0.0.1:{info['port']}"
            table.add_row(service_name, 
                          str(info['port']), 
                          str(info['pid']), 
                          url, 
                          str(info['local']), 
                          str(info['logs']) )
        
        console.print(table)

def start_local(service_name, port, app_type, app_folder, log_file):
    """
    Starts a local Flask or FastAPI VAC app.

    Args:
        service_name (str): Name of the service.
        port (int): Port to run the local app on.
        app_type (str): Type of the app ('flask' or 'fastapi').
        app_folder (str): Folder containing the app.
    """
    proxies = clean_proxy_list()

    if service_name in proxies:
        console.print(f"Local VAC app [bold orange]'{service_name}'[/bold orange] is already running on port {proxies[service_name]['port']}.")
        return

    if app_type == 'flask':
        command = ["gunicorn", "--bind", f"0.0.0.0:{port}", "--workers", "4", "app:app"]
    elif app_type == 'fastapi':
        command = ["uvicorn", "app:app", f"--port={port}"]
    else:
        print(f"[bold red]Unknown app type: {app_type}[/bold red]")
        return

    if log_file:
        log_file_path = os.path.join(app_folder, f"{service_name}_log.txt")
        with open(log_file_path, 'w') as logf:
            process = subprocess.Popen(command, cwd=app_folder, stdout=logf, stderr=logf, preexec_fn=os.setpgrp)
    else:
        log_file_path = "No log file specified"
        with open(os.devnull, 'a') as devnull:
            process = subprocess.Popen(command, cwd=app_folder, stdout=devnull, stderr=devnull, preexec_fn=os.setpgrp)

    proxies[service_name] = {
        "pid": process.pid,
        "port": port,
        "local": f"{app_folder}/app.py - {app_type}",
        "logs": log_file_path
    }
    save_proxies(proxies)
    
    console.print(f"Local app [bold orange]'{service_name}'[/bold orange] started on port {port}")
    list_proxies()

    return f"http://127.0.0.1:{port}"

def setup_proxy_subparser(subparsers):
    """
    Sets up an argparse subparser for the 'proxy' command.

    Args:
        subparsers: The subparsers object from argparse.ArgumentParser().
    """
    
    proxy_parser = subparsers.add_parser('proxy', help='Set up or stop a proxy to the VAC Cloud Run services')
    proxy_subparsers = proxy_parser.add_subparsers(dest='proxy_command', required=True)

    start_parser = proxy_subparsers.add_parser('start', help='Start the proxy to the VAC Cloud Run service')
    start_parser.add_argument('service_name', help='Name of the Cloud Run service.')
    start_parser.add_argument('--port', type=int, help='Port to run the proxy on. Auto-assigns if not provided.')
    start_parser.add_argument('--local', action='store_true', help='Run the service locally instead of proxying to Cloud Run.')
    start_parser.add_argument('--app-type', choices=['flask', 'fastapi'], help='If local, type of the local app (flask or fastapi).')
    start_parser.add_argument('--app-folder', default=".", help='If local, folder containing the local app.py')
    start_parser.add_argument('--log-file', action='store_true', help='Whether to create a file containing proxy logs.')
    start_parser.set_defaults(func=lambda args: start_proxy(args.service_name, 
                                                            args.region, 
                                                            args.project, 
                                                            args.port, 
                                                            args.local, 
                                                            args.app_type, 
                                                            args.app_folder,
                                                            args.log_file))
    
    stop_parser = proxy_subparsers.add_parser('stop', help='Stop the proxy to the Cloud Run service. Backup: `kill -9 PID`')
    stop_parser.add_argument('service_name', help='Name of the Cloud Run service.')
    stop_parser.set_defaults(func=lambda args: stop_proxy(args.service_name))

    list_parser = proxy_subparsers.add_parser('list', help='List all running proxies. Examine port via `lsof-i :PORT`')
    list_parser.set_defaults(func=lambda args: list_proxies())

    stop_all_parser = proxy_subparsers.add_parser('stop-all', help='Stop all running proxies.')
    stop_all_parser.set_defaults(func=lambda args: stop_all_proxies())






import traceback

from ..custom_logging import log
from ..utils import ConfigManager
from .safety import genai_safety

from typing import TYPE_CHECKING, Union

import json
from collections import deque

try:
    import google.generativeai as genai
    import proto
    from google.generativeai.types import RequestOptions, GenerateContentResponse
    from google.api_core import retry
    from google.generativeai import ChatSession
    from google.api_core.exceptions import RetryError
except ImportError:
    genai = None
    ChatSession = None

from .images import extract_gs_images_and_genai_upload

if TYPE_CHECKING:
    from google.generativeai.protos import Part
    from google.generativeai import ChatSession



class GenAIFunctionProcessor:
    """
    A generic class for processing function calls from google.generativeai function calling models.

    This class provides a framework for handling multiple function calls in responses
    from generative AI systems. Users of this class should subclass it and provide
    their own implementation of the `construct_tools` method, which returns a dictionary
    of function names mapped to their implementations.

    Attributes:
        config (ConfigManager): Configuration manager instance. Reach values via self.config within your own construct_tools() method
        funcs (dict): A dictionary of function names mapped to their implementations.

    Example usage:

    ```python
    class AlloyDBFunctionProcessor(GenAIFunctionProcessor):
        def construct_tools(self) -> dict:
            pass

    config = ConfigManager()
    alloydb_processor = AlloyDBFunctionProcessor(config)

    results = alloydb_processor.process_funcs(full_response)

    alloydb_model = alloydb_processor.get_model(
        model_name="gemini-1.5-pro",
        system_instruction="You are a helpful AlloyDB agent that helps users search and extract documents from the database."
    )
    ```
    """

    def __init__(self, config: ConfigManager=None, model_name=None, trace=None, parent_observation_id=None):
        """
        Initializes the GenAIFunctionProcessor with the given configuration.

        Args:
            config (ConfigManager): The configuration manager instance.
            model_name (str): The name of the model
        """
        if not genai:
            raise ImportError("import google.generativeai as genai is required, import via `pip install sunholo[gcp]`")
        
        self.config = config
        self.funcs = self.construct_tools()

        # Add default 'decide_to_go_on' if not provided in construct_tools
        if 'decide_to_go_on' not in self.funcs:
            self.funcs['decide_to_go_on'] = self.decide_to_go_on

        self.model_name = "gemini-1.5-flash"
        if config:
            self.model_name = config.vacConfig("model") if config.vacConfig("llm") == "vertex" else "gemini-1.5-flash"
        
        if model_name:
            log.info(f"Overriding agent model name {self.model_name} with model {model_name}")
            self.model_name = model_name
        
        self.trace = trace
        self.parent_observation_id = parent_observation_id
        
        self.last_api_requests_and_responses = []
        self._validate_functions()

    def construct_tools(self) -> dict:
        """
        Constructs a dictionary of tools (functions) specific to the application.

        This method should be overridden in subclasses to provide the specific
        function implementations required for the application.
        
        Note: All functions need arguments to avoid errors.

        Returns:
            dict: A dictionary where keys are function names and values are function objects

        Raises:
            NotImplementedError: If the method is not overridden in a subclass.
        """
        raise NotImplementedError("Subclasses must implement this method to return a dictionary of functions.")

    def _validate_functions(self):
        """
        Validates that all functions in the `funcs` dictionary have docstrings.

        This method checks each function in the `funcs` dictionary to ensure it has
        a docstring. If a function is missing a docstring, an error is logged, and
        a `ValueError` is raised.

        Raises:
            ValueError: If any function is missing a docstring.
        """
        for func_name, func in self.funcs.items():
            if not func.__doc__:
                log.error(f"Function {func_name} is missing a docstring.")
                raise ValueError(f"Function {func_name} must have a docstring to be used as a genai tool.")
    

    def parse_as_parts(self, api_requests_and_responses=[]):
        if not api_requests_and_responses and not self.last_api_requests_and_responses:
            log.info("No api_requests_and_responses found to parse to parts")
            return None

        if api_requests_and_responses:
            self.last_api_requests_and_responses = api_requests_and_responses
        
        from google.generativeai.protos import Part

        work_on = api_requests_and_responses or self.last_api_requests_and_responses
        parts = []
        for part in work_on:
            try:
                parts.append(
                    Part(
                        function_response=genai.protos.FunctionResponse(
                            name=part[0],
                            response={"result": part[2], "args": json.dumps(part[1])}
                        )
                    )
                )
            except Exception as err:
                parts.append(
                    Part(
                        function_response=genai.protos.FunctionResponse(
                            name=part[0],
                            response={"result": f"ERROR: {str(err)}"}
                        )
                    )
                )                

        return parts

    def check_function_result(self, function_name, target_value, api_requests_and_responses=[]):
        """
        Checks if a specific function result in the api_requests_and_responses contains a certain value.

        Args:
            function_name (str): The name of the function to check.
            target_value: The value to look for in the function result.
            api_requests_and_responses (list, optional): List of function call results to check. 
                                                        If not provided, the method will use `self.last_api_requests_and_responses`.

        Returns:
            bool: True if the target_value is found in the specified function's result, otherwise False.
        """
        if not api_requests_and_responses:
            api_requests_and_responses = self.last_api_requests_and_responses

        if not api_requests_and_responses:
            log.info("No api_requests_and_responses found to check.")
            return False

        for part in api_requests_and_responses:
            func_name = part[0]
            result = part[2]

            if func_name == function_name:
                # Try to decode the result if it's a string
                if isinstance(result, str):
                    try:
                        result = json.loads(result)
                    except json.JSONDecodeError:
                        log.warning(f"Failed to decode JSON result for function {function_name}: {result}")
                        continue  # Skip this result if decoding fails

                normalized_target_value = {k: v for k, v in target_value.items()} if isinstance(target_value, dict) else target_value
                log.info(f"{normalized_target_value=} {result=}") 

                if isinstance(result, dict) and isinstance(normalized_target_value, dict):
                    for key, expected_value in normalized_target_value.items():
                        if key in result and result[key] == expected_value:
                            log.info(f"The key '{key}' has the expected value in both dictionaries.")
                            return True
                    return False
                elif result == normalized_target_value:
                    log.info(f"Target value '{target_value}' found in the result of function '{function_name}'.")
                    return True

        log.info(f"Target value '{target_value}' not found in the result of function '{function_name}'.")
        return False
    
    def parse_as_string(self, api_requests_and_responses=[]):
        if not api_requests_and_responses and not self.last_api_requests_and_responses:
            log.info("No api_requests_and_response found to parse to string")
            return None
        
        if api_requests_and_responses:
            self.last_api_requests_and_responses = api_requests_and_responses
        
        work_on = api_requests_and_responses or self.last_api_requests_and_responses
        strings = []
        for part in work_on:
            strings.append(
                f"function tool {part[0]} was called with arguments: {part[1]} and got this result:\n"
                f"<{part[0]}_result>{part[2]}</{part[0]}_result>"
            )

        return strings

    def process_funcs(self, full_response, output_parts=True, loop_span=None) -> Union[list['Part'], str]:
        """
        Processes the functions based on the full_response from the generative model.

        This method iterates through each part of the response, extracts function
        calls and their parameters, and executes the corresponding functions defined
        in the `funcs` dictionary.

        Args:
            full_response: The response object containing function calls.
            output_parts (bool): Indicates whether to return structured parts or plain strings.

        Returns:
            list[Part] | str: A list of Part objects or a formatted string with the results.

        Example usage:
        ```python
        results = alloydb_processor.process_funcs(full_response)
        ```
        """
        api_requests_and_responses = []


        if not full_response:
            log.info("No response was found to process")
            return api_requests_and_responses

        # Loop through each part in the response to handle multiple function calls
        #TODO: async
        if not full_response.candidates or len(full_response.candidates) == 0:
            log.error("No candidates found in the response. The response might have failed.")
            return "No candidates available in the response. Please check your query or try again."

        for part in full_response.candidates[0].content.parts:
            if fn := part.function_call:
                # Extract parameters for the function call
                function_name = fn.name
                
                # Handle empty parameters
                if fn.args is None or not fn.args:
                    params_obj = {}
                else:
                    params_obj = {key: val for key, val in fn.args.items()}

                params = ', '.join(f'{key}={val}' for key, val in params_obj.items())
                log.info(f"Executing {function_name} with params {params} (Total Characters: {len(params)})")
                if len(params)>8000:
                    log.warning(f"Total parameters are over 8000 characters - it may not work properly: {params[:10000]}....[{len(params)}]")

                fn_span = loop_span.span(name=function_name, input=params_obj) if loop_span else None

                # Check if the function is in our dictionary of available functions
                if function_name in self.funcs:
                    fn_exec = self.funcs[function_name]
                    try:
                        if not isinstance(fn_exec, genai.protos.FunctionDeclaration):
                            # Execute the function with the provided parameters
                            result = fn_exec(**params_obj)
                            log.info(f"Got result from {function_name}: {result} of type: {type(result)}")
                            #TODO: return images
                        else:
                            fn_result = type(fn).to_dict(fn)
                            result = fn_result.get("result")
                            if not result:
                                log.warning("No result found for {function_name}")
                            log.info(f"No execution of {function_name} as a FunctionDeclatation object, just returning args {result}")
                    except Exception as err:
                        error_message = f"Error in {function_name}: {str(err)}"
                        traceback_details = traceback.format_exc()
                        log.warning(f"{error_message}\nTraceback: {traceback_details}")
                        result = [error_message] #traceback uses too many tokens
                    clean_result = self.remove_invisible_characters(result)
                    api_requests_and_responses.append(
                        [function_name, params, clean_result]
                    )
                    fn_span.end(output=clean_result) if fn_span else None
                else:
                    msg = f"Function {function_name} is not recognized"
                    log.error(msg)
                    fn_span.end(output=msg) if fn_span else None

        log.info(f"{api_requests_and_responses=}")
        self.last_api_requests_and_responses = api_requests_and_responses

        if output_parts:
            return self.parse_as_parts()
        
        return self.parse_as_string()
    
    def tool_config_setting(self, mode:str):
        from google.generativeai.types import content_types

        fns = list(self.funcs.keys())

        if fns and mode == "any":
            return content_types.to_tool_config(
                    {"function_calling_config": {"mode": mode, "allowed_function_names": fns}}
                )
        else:
            return content_types.to_tool_config(
                {"function_calling_config": {"mode": mode}}
            )

    def get_model(
            self, 
            system_instruction: str, 
            generation_config=None, 
            model_name: str=None,
            tool_config: str="auto"):
        """
        Constructs and returns the generative AI model configured with the tools.

        This method creates a generative AI model using the tools defined in the
        `funcs` dictionary and the provided configuration options.

        Args:
            model_name (str): The name of the model to use.
            system_instruction (str): Instructions for the AI system.
            generation_config (dict, optional): Configuration for generation, such as temperature.
            tool_config (str, optional): Configuration for tool behaviour: 'auto' it decides, 'none' no tools, 'any' always use tools

        Returns:
            GenerativeModel: An instance of the GenerativeModel configured with the provided tools.

        Example usage:

        ```python
        alloydb_model = alloydb_processor.get_model(
            model_name="gemini-1.5-pro",
            system_instruction="You are a helpful AlloyDB agent that helps users search and extract documents from the database."
        )
        ```
        """
        if generation_config is None:
            generation_config = {
                "temperature": 0.1,
                "max_output_tokens": 8000,
            }

        # Extract the functions from the dictionary to pass into the model
        tools = list(self.funcs.values())

        try:
            model = genai.GenerativeModel(
                model_name=model_name or self.model_name,
                tools=tools,
                tool_config=self.tool_config_setting(tool_config),
                generation_config=generation_config,
                safety_settings=genai_safety(),
                system_instruction=system_instruction,
            )
            return model
        except Exception as err:
            log.error(f"Error initializing model: {str(err)}")
            return None
    
    def remove_invisible_characters(self, string):

        if not isinstance(string, str):
            return string
        
        clean = string.encode('utf-8', errors='replace').decode('unicode_escape')
        log.info(f"Cleaning:\n{string[:100]}\n > to >\n{clean[:100]}")
        
        return clean
    
    def convert_composite_to_native(self, value):
        """
        Recursively converts a proto MapComposite or RepeatedComposite object to native Python types.

        Args:
            value: The proto object, which could be a MapComposite, RepeatedComposite, or a primitive.

        Returns:
            The equivalent Python dictionary, list, or primitive type.
        """
        if isinstance(value, proto.marshal.collections.maps.MapComposite):
            # Convert MapComposite to a dictionary, recursively processing its values
            return {key: self.convert_composite_to_native(val) for key, val in value.items()}
        elif isinstance(value, proto.marshal.collections.repeated.RepeatedComposite):
            # Convert RepeatedComposite to a list, recursively processing its elements
            return [self.convert_composite_to_native(item) for item in value]
        else:
            # If it's a primitive value, return it as is
            return value

    def run_agent_loop(self, chat:ChatSession, content:list, callback=None, guardrail_max=10, loop_return=3): # type: ignore
        """
        Runs the agent loop, sending messages to the orchestrator, processing responses, and executing functions.

        Args:
            chat: The chat object for interaction with the orchestrator.
            content: The initial content to send to the agent.
            callback: The callback object for handling intermediate responses. If not supplied will use self.IOCallback()
            guardrail_max (int): The maximum number of iterations for the loop.
            loop_return (int): The number of last loop iterations to return. Default 3 will return last 3 iterations. If loop_return > guardrail_max then all iterations are returned.

        Returns:
            tuple: (big_text, usage_metadata) from the loop execution.
        """
        if not callback:
            callback = self.IOCallback()
        guardrail = 0
        big_result = []
        usage_metadata = {
            "prompt_token_count": 0,
            "candidates_token_count": 0,
            "total_token_count": 0
        }
        functions_called =[]
        function_results = []
        # Initialize token queue to ensure sequential processing
        token_queue = deque()

        span = self.trace.span(
            name=f"GenAIFunctionProcesser_{self.__class__.__name__}",
            parent_observation_id=self.parent_observation_id,
            input = {'content': content},
        ) if self.trace else None

        while guardrail < guardrail_max:

            token_queue.append(f"\n----Loop [{guardrail}] Start------\nFunctions: {list(self.funcs.keys())}\n")

            content_parse = ""
            for i, chunk in enumerate(content):
                content_parse += f"\n - {i}) {chunk}"
            content_parse += f"\n== End input content for loop [{guardrail}] =="

            log.info(f"== Start input content for loop [{guardrail}]\n ## Content: {content_parse}")
            this_text = ""  # reset for this loop
            response = None
            executed_responses = []
            loop_span = span.span(
                name=f"loop_{guardrail}",
                model=self.model_name,
                input = {'content': content},
            ) if span else None

            try:
                token_queue.append("\n= Calling Agent =\n")

                gen = loop_span.generation(
                    name=f"loop_{guardrail}",
                    model=self.model_name,
                    input = {'content': content},
                ) if loop_span else None

                response: GenerateContentResponse = chat.send_message(content, request_options=RequestOptions(
                                        retry=retry.Retry(
                                            initial=1, 
                                            multiplier=2, 
                                            maximum=10, 
                                            timeout=60
                                        )
                                       ))
            except RetryError as err:
                msg = f"Retry error - lets try again if its occured less than twice: {str(err)}"
                log.warning(msg)
                token_queue.append(msg)
                this_text += msg
                
            except Exception as e:
                msg = f"Error sending {content} to model: {str(e)}"
                if str(e) == "finish_reason: 10":
                    msg = "The Gemini API does not work with this input - you need to try something else. Error is: finish_reason: 10"
                log.error(msg + f"{traceback.format_exc()}")
                token_queue.append(msg)
                this_text += msg

            if response:
                loop_metadata = response.usage_metadata
                if loop_metadata:
                    usage_metadata = {
                        "prompt_token_count": usage_metadata["prompt_token_count"] + (loop_metadata.prompt_token_count or 0),
                        "candidates_token_count": usage_metadata["candidates_token_count"] + (loop_metadata.candidates_token_count or 0),
                        "total_token_count": usage_metadata["total_token_count"] + (loop_metadata.total_token_count or 0),
                    }
                    token_queue.append((
                        "\n-- Agent response -- " 
                        f"Loop tokens: [{loop_metadata.prompt_token_count}]/[{usage_metadata['prompt_token_count']}] "
                        f"Session tokens: [{loop_metadata.total_token_count}]/[{usage_metadata['total_token_count']}] \n"
                    ))
                loop_metadata = None
                gen.end(output=response.to_dict()) if gen else None
            else:
                gen.end(output="No response received") if gen else None

            if not response:
                response = []
            for chunk in response:
                if not chunk:
                    continue

                log.debug(f"[{guardrail}] {chunk=}")
                try:
                    if hasattr(chunk, 'text') and isinstance(chunk.text, str):
                        token = chunk.text
                        token_queue.append(token)
                        this_text += token
                    else:
                        log.info("skipping chunk with no text")
                    
                except ValueError as err:
                    token_queue.append(f"{str(err)} for {chunk=}")
            try:
                executed_responses = self.process_funcs(response, loop_span=loop_span) 
            except Exception as err:
                log.error(f"Error in executions: {str(err)}")
                token_queue.append(f"{str(err)} for {response=}")

            log.info(f"[{guardrail}] {executed_responses=}")

            if executed_responses:  
                token_queue.append("\n-- Agent Actions:\n")
                fn_exec = loop_span.span(name="function_actions", input=executed_responses) if loop_span else None
                for executed_response in executed_responses:
                    token = ""
                    fn = executed_response.function_response.name
                    fn_args = executed_response.function_response.response.get("args")
                    fn_result = executed_response.function_response.response["result"]
                    fn_log = f"{fn}({fn_args})"
                    log.info(fn_log)
                    functions_called.append(fn_log)
                    function_results.append(fn_result)
                    token_queue.append(f"\n-- {fn_log} ...executing...\n") if fn != "decide_to_go_on" else ""
                    while token_queue:
                        token = token_queue.popleft()
                        callback.on_llm_new_token(token=token)

                    log.info(f"{fn_log} created a result={type(fn_result)=}")
                    fn_exec_one = fn_exec.span(name=fn, input=fn_args) if fn_exec else None
                    
                    fn_result_json = None
                    # Convert MapComposite to a standard Python dictionary
                    if isinstance(fn_result, proto.marshal.collections.maps.MapComposite):
                        fn_result_json = self.convert_composite_to_native(fn_result)
                    elif isinstance(fn_result, proto.marshal.collections.repeated.RepeatedComposite):
                        fn_result = self.convert_composite_to_native(fn_result)
                    elif isinstance(fn_result, dict):
                        fn_result_json = fn_result
                    elif isinstance(fn_result, str):
                        try:
                            if isinstance(fn_result_json, str):
                                fn_result_json = json.loads(fn_result_json)
                        except json.JSONDecodeError:
                            log.warning(f"{fn_result} was not JSON decoded")
                        except Exception as err:
                            log.warning(f"{fn_result} was not json decoded due to unknown exception: {str(err)} {traceback.format_exc()}")
                    else:
                        log.warning(f"Unrecognised type for {fn_log}: {type(fn_result)}")
                    
                    # should be a string or a dict by now
                    log.info(f"Processed {fn_log} to {fn_result_json=} type: {type(fn_result_json)}")
                    
                    if fn == "decide_to_go_on":
                        log.info(f"{fn_result_json=} {type(fn_result)}")
                        if fn_result_json:
                            token = f"\n{'STOPPING' if not fn_result_json.get('go_on') else 'CONTINUE'}: {fn_result_json.get('chat_summary')}\n"
                        else:
                            log.warning(f"{fn_result_json} did not work for decide_to_go_on")
                            token = f"Error calling decide_to_go_on with {fn_result=}\n"
                    else:
                        token = f"--- {fn_log} result --- \n"
                        # if json dict we look for keys to extract
                        if fn_result_json:
                            log.info(f"{fn_result_json} dict parsing")
                            if fn_result_json.get('stdout'):
                                text = fn_result_json.get('stdout')
                                token += self.remove_invisible_characters(text)
                            if fn_result_json.get('stderr'):
                                text = fn_result_json.get('stdout')
                                token += self.remove_invisible_characters(text)
                            # If neither 'stdout' nor 'stderr' is present, dump the entire JSON
                            if 'stdout' not in fn_result_json and 'stderr' not in fn_result_json:
                                log.info(f"No recognised keys ('stdout' or 'stderr') in dict: {fn_result_json=} - dumping it all")
                                token += f"{json.dumps(fn_result_json, indent=2)}\n"  # Added `indent=2` for readability
                        else:
                            # probably a string, just return it
                            log.info(f"{fn_result_json} non-dict (String?) parsing")
                            token += f"{self.remove_invisible_characters(fn_result)}\n--- end ---\n"
                    
                    this_text += token
                    token_queue.append(token)
                    fn_exec_one.end(output=token) if fn_exec_one else None
                fn_exec.end(output=this_text) if fn_exec else None        

            else:
                token = "\nNo function executions were found\n"
                token_queue.append(token)
                this_text += token

            if this_text:
                #content.append(f"Agent: {this_text}")    
                # update content relying on gemini chat history instead
                content = [f"Agent:\n{this_text}"]
                # if text includes gs:// try to download it
                image_uploads = extract_gs_images_and_genai_upload(this_text)
                if image_uploads:
                    for img in image_uploads:
                        log.info(f"Adding {img=}")
                        content.append(img)
                        content.append(f"{img.name} was created by agent and added")
                log.info(f"[{guardrail}] Updated content:\n{this_text}")
                big_result.append(this_text)
            else:
                log.warning(f"[{guardrail}] No content created this loop")
                content = [f"Agent: ERROR - No response was found for loop [{guardrail}]"]

            token_queue.append(f"\n----Loop [{guardrail}] End------\n{usage_metadata}\n----------------------")
            loop_span.end(output=content, metadata=loop_metadata) if loop_span else None

            go_on_check = self.check_function_result("decide_to_go_on", {"go_on": False})
            if go_on_check:
                log.info("Breaking agent loop")
                break
            
            while token_queue:
                token = token_queue.popleft()
                callback.on_llm_new_token(token=token)
            
            guardrail += 1
            if guardrail > guardrail_max:
                log.warning(f"Guardrail kicked in, more than {guardrail_max} loops")
                break
        
        while token_queue:
            token = token_queue.popleft()
            callback.on_llm_new_token(token=token)
        
        usage_metadata["functions_called"] = functions_called

        big_text = "\n".join(big_result[-loop_return:])
        span.end(output=big_text, metadata=usage_metadata) if span else None

        return big_text, usage_metadata

    class IOCallback:
        """
        This is a default callback that will print to console any tokens it recieves.
        """
        def on_llm_new_token(self, token:str):
            print(token)
        def on_llm_end(self, response):
            print(f"\nFull response: \n{response}")

    @staticmethod
    def decide_to_go_on(go_on: bool, chat_summary: str) -> dict:
        """
        Examine the chat history.  If the answer to the user's question has been answered, then go_on=False.
        If the chat history indicates the answer is still being looked for, then go_on=True.
        If there is no chat history, then go_on=True.
        If there is an error that can't be corrected or solved by you, then go_on=False.
        If there is an error but you think you can solve it by correcting your function arguments (such as an incorrect source), then go_on=True
        If you want to ask the user a question or for some more feedback, then go_on=False.  
        Avoid asking the user if you suspect you can solve it yourself with the functions at your disposal - you get top marks if you solve it yourself without help.
        When calling, please also add a chat summary of why you think the function should  be called to end.
        
        Args:
            go_on: boolean Whether to continue searching for an answer
            chat_summary: string A brief explanation on why go_on is TRUE or FALSE
        
        Returns:
            boolean: True to carry on, False to continue
        """
        return {"go_on": go_on, "chat_summary": chat_summary}
    


try:
    from google.cloud import texttospeech
except ImportError:
    texttospeech = None
try:
    import sounddevice as sd
except ImportError:
    sd = None
except OSError:
    sd = None

try:
    import numpy as np
except ImportError:
    np = None

try:
    from rich import console
    console = console.Console()
except ImportError:
    console = None

from ..custom_logging import log
import queue
import threading
import time
from concurrent.futures import ThreadPoolExecutor
import io
import wave

import argparse
import json
from typing import Optional
from pathlib import Path
import sys

class StreamingTTS:
    """
    # Example usage
    def sample_text_stream():
        sentences = [
            "Hello, this is a test of streaming text to speech.",
            "Each sentence will be converted to audio separately.",
            "This allows for lower latency in long-form text to speech conversion."
        ]
        for sentence in sentences:
            yield sentence
            time.sleep(0.5)  # Simulate delay between text chunks

    # Initialize and run
    tts = StreamingTTS()
    tts.process_text_stream(sample_text_stream())    
    """
    def __init__(self):
        if texttospeech is None or sd is None or np is None:
            raise ImportError(f"StreamingTTS requires imports via pip install sunholo[tts] - {texttospeech=} {sd=} {np=}")
        
        log.info("Initializing StreamingTTS...")
        self.client = texttospeech.TextToSpeechClient()
        self.audio_queue = queue.Queue()
        self.is_playing = False
        self.sample_rate = 24000  # Google's default sample rate
        self.language_code = "en-GB"
        self.voice_gender = texttospeech.SsmlVoiceGender.NEUTRAL
        self.voice_name = "en-GB-Journey-D"
        # Audio processing parameters
        # Separate fade durations for playback and file saving
        self.playback_fade_duration = 0.05  # 50ms fade for real-time playback
        self.file_fade_duration = 0.01      # 10ms fade for file saving
        self.stream = None
        self._initialize_audio_device()

    def set_voice(self, voice_name: str):
        """
        Set the language for text-to-speech conversion.
        
        Args:
            language_code: Language code in BCP-47 format (e.g., 'en-US', 'es-ES', 'fr-FR')
        """
        log.info(f"Setting voice to {voice_name}")
        self.voice_name = voice_name

    def set_language(self, language_code: str):
        """
        Set the language for text-to-speech conversion.
        
        Args:
            language_code: Language code in BCP-47 format (e.g., 'en-US', 'es-ES', 'fr-FR')
        """
        log.info(f"Setting language to {language_code}")
        self.language_code = language_code
        
    def set_voice_gender(self, gender: str):
        """
        Set the voice gender for text-to-speech conversion.
        
        Args:
            gender: One of 'NEUTRAL', 'MALE', or 'FEMALE'
        """
        gender_map = {
            'NEUTRAL': texttospeech.SsmlVoiceGender.NEUTRAL,
            'MALE': texttospeech.SsmlVoiceGender.MALE,
            'FEMALE': texttospeech.SsmlVoiceGender.FEMALE
        }
        
        if gender not in gender_map:
            raise ValueError(f"Invalid gender '{gender}'. Must be one of: {', '.join(gender_map.keys())}")
        
        log.info(f"Setting voice gender to {gender}")
        self.voice_gender = gender_map[gender]

    def text_to_audio(self, text):
        """Convert text chunk to audio bytes using Google Cloud TTS."""
        log.info(f"TTS: {text=}")
        synthesis_input = texttospeech.SynthesisInput(text=text)
        
        voice = texttospeech.VoiceSelectionParams(
            language_code=self.language_code,
            ssml_gender=self.voice_gender,
            name=self.voice_name
        )
        
        audio_config = texttospeech.AudioConfig(
            audio_encoding=texttospeech.AudioEncoding.LINEAR16,
            sample_rate_hertz=self.sample_rate
        )
        
        response = self.client.synthesize_speech(
            input=synthesis_input,
            voice=voice,
            audio_config=audio_config
        )
        log.info("Got response from TTS")
        
        # Convert audio bytes to numpy array for playback
        audio_np = np.frombuffer(response.audio_content, dtype=np.int16)
        return audio_np

    def generate_audio_stream(self, text):
        """
        Generate a stream of audio data from a text chunk.
        Returns audio in WAV format for streaming.
        
        Args:
            text (str): Text to convert to speech
            
        Yields:
            bytes: WAV-formatted audio data
        """
        try:
            # Convert text to audio using existing method
            audio_chunk = self.text_to_audio(text)
            
            # Process audio chunk with fading
            processed_chunk = self._apply_fade(
                audio_chunk,
                fade_duration=self.file_fade_duration,
                fade_in=True,
                fade_out=True
            )
            
            # Convert to WAV format
            wav_buffer = io.BytesIO()
            with wave.open(wav_buffer, 'wb') as wav_file:
                wav_file.setnchannels(1)
                wav_file.setsampwidth(2)
                wav_file.setframerate(self.sample_rate)
                wav_file.writeframes(processed_chunk.tobytes())
            
            yield wav_buffer.getvalue()
            
        except Exception as e:
            log.error(f"Error generating audio stream: {e}")
            yield b''
    
    def _initialize_audio_device(self):
        """Initialize audio device with proper settings."""
        try:
            # Set default device settings
            sd.default.samplerate = self.sample_rate
            sd.default.channels = 1
            sd.default.dtype = np.int16
            
            # Initialize persistent output stream
            self.stream = sd.OutputStream(
                samplerate=self.sample_rate,
                channels=1,
                dtype=np.int16,
                latency='low'
            )
            self.stream.start()
            
            log.info("Audio device initialized successfully")
        except Exception as e:
            log.error(f"Error initializing audio device: {e}")
            raise

    def _make_fade(self, length: int, fade_type: str='l') -> np.ndarray:
        """Generate a fade curve of specified length and type."""
        fade = np.arange(length, dtype=np.float32) / length
        
        if fade_type == 't':  # triangle
            pass
        elif fade_type == 'q':  # quarter of sinewave
            fade = np.sin(fade * np.pi / 2)
        elif fade_type == 'h':  # half of sinewave
            fade = (1 - np.cos(fade * np.pi)) / 2
        elif fade_type == 'l':  # logarithmic
            fade = np.power(0.1, (1 - fade) * 5)
        elif fade_type == 'p':  # inverted parabola
            fade = (1 - (1 - fade)**2)
        else:
            raise ValueError(f"Unknown fade type {fade_type!r}")
        
        return fade
    
    def _apply_fade(self, audio: np.ndarray, fade_duration: float, fade_in: bool = True, fade_out: bool = True) -> np.ndarray:
        """Apply fade in/out to audio with specified duration."""
        if audio.ndim != 1:
            raise ValueError("Audio must be 1-dimensional")
        
        fade_length = int(fade_duration * self.sample_rate)
        audio = audio.astype(np.float32)
        
        if fade_in:
            fade_in_curve = self._make_fade(fade_length, 'l')
            audio[:fade_length] *= fade_in_curve
            
        if fade_out:
            fade_out_curve = self._make_fade(fade_length, 'l')
            audio[-fade_length:] *= fade_out_curve[::-1]
        
        return audio.astype(np.int16)

    
    def _play_audio_chunk(self, audio_chunk: np.ndarray, is_final_chunk: bool = False):
        """Play a single audio chunk with proper device handling."""
        try:
            # Add longer padding for the final chunk
            padding_duration = 0.1 if is_final_chunk else 0.02
            padding = np.zeros(int(padding_duration * self.sample_rate), dtype=np.int16)
            
            if is_final_chunk:
                # For final chunk, add extra padding and longer fade
                audio_with_padding = np.concatenate([
                    padding, 
                    audio_chunk, 
                    padding,
                    np.zeros(int(0.2 * self.sample_rate), dtype=np.int16)  # Extra tail padding
                ])
                fade_duration = self.playback_fade_duration * 2  # Longer fade for end
            else:
                audio_with_padding = np.concatenate([padding, audio_chunk, padding])
                fade_duration = self.playback_fade_duration
            
            processed_audio = self._apply_fade(
                audio_with_padding,
                fade_duration=fade_duration,
                fade_in=True,
                fade_out=True
            )
            
            if self.stream and self.stream.active:
                self.stream.write(processed_audio)
                if is_final_chunk:
                    # Write a small buffer of silence at the end
                    final_silence = np.zeros(int(0.1 * self.sample_rate), dtype=np.int16)
                    self.stream.write(final_silence)
            else:
                with sd.OutputStream(
                    samplerate=self.sample_rate,
                    channels=1,
                    dtype=np.int16,
                    latency='low'
                ) as temp_stream:
                    temp_stream.write(processed_audio)
                    if is_final_chunk:
                        temp_stream.write(np.zeros(int(0.1 * self.sample_rate), dtype=np.int16))
            
        except Exception as e:
            log.error(f"Error during audio playback: {e}")
            raise
    
    def audio_player(self):
        """Continuously play audio chunks from the queue."""
        log.info("Audio player started")
        try:
            while self.is_playing or not self.audio_queue.empty():
                if not self.audio_queue.empty():
                    audio_chunk = self.audio_queue.get()
                    self._play_audio_chunk(audio_chunk)
                time.sleep(0.005)  # Reduced sleep time for more responsive playback
        finally:
            # Ensure stream is properly closed
            if self.stream and self.stream.active:
                self.stream.stop()
                self.stream.close()
                self.stream = None

    def __del__(self):
        """Cleanup method to ensure stream is closed."""
        if hasattr(self, 'stream') and self.stream and self.stream.active:
            # Write a small silence buffer before closing
            final_silence = np.zeros(int(0.1 * self.sample_rate), dtype=np.int16)
            try:
                self.stream.write(final_silence)
                time.sleep(0.1)  # Let the final audio finish playing
            except Exception:
                pass  # Ignore errors during cleanup
            self.stream.stop()
            self.stream.close()
    
    def process_text_stream(self, text_generator):
        """Process incoming text stream and convert to audio."""
        self.is_playing = True
        
        # Start audio playback thread
        player_thread = threading.Thread(target=self.audio_player)
        player_thread.start()
        
        try:
            # Process text chunks in parallel
            with ThreadPoolExecutor(max_workers=3) as executor:
                futures = []
                for text_chunk in text_generator:
                    future = executor.submit(self.text_to_audio, text_chunk)
                    futures.append(future)
                
                # Process results as they complete
                for future in futures:
                    audio_chunk = future.result()
                    self.audio_queue.put(audio_chunk)
        finally:
            self.is_playing = False
            player_thread.join()
    
    def save_to_file(self, text_generator, output_path):
        """Save the audio to a WAV file with minimal fading."""
        import wave
        
        all_audio = []
        for text_chunk in text_generator:
            audio_chunk = self.text_to_audio(text_chunk)
            # Use shorter fade duration for file saving
            processed_chunk = self._apply_fade(
                audio_chunk,
                fade_duration=self.file_fade_duration
            )
            all_audio.append(processed_chunk)
        
        # Add minimal silence between chunks
        silence = np.zeros(int(0.05 * self.sample_rate), dtype=np.int16)
        final_audio = silence
        
        for i, chunk in enumerate(all_audio):
            if i == len(all_audio) - 1:
                # For the last chunk, use a slightly longer fade out
                chunk = self._apply_fade(
                    chunk,
                    fade_duration=self.file_fade_duration * 2,
                    fade_in=False,
                    fade_out=True
                )
            final_audio = np.concatenate([final_audio, chunk, silence])
        
        with wave.open(output_path, 'wb') as wav_file:
            wav_file.setnchannels(1)
            wav_file.setsampwidth(2)
            wav_file.setframerate(self.sample_rate)
            wav_file.writeframes(final_audio.tobytes())

def tts_command(args):
    """
    Executes the TTS command based on parsed arguments.
    
    Args:
        args: The parsed command-line arguments.
    """
    if console is None:
        raise ImportError("Need cli tools to use TTS commands - install via `pip install sunholo[cli,tts]`")
    
    from rich.panel import Panel

    def text_generator(input_source: str, is_file: bool = False):
        """Generate text from either a file or direct input."""
        if is_file:
            try:
                with open(input_source, 'r') as f:
                    for line in f:
                        line = line.strip()
                        if line:  # Skip empty lines
                            yield line
            except FileNotFoundError:
                console.print(f"Error: The input file '{input_source}' was not found.")
                sys.exit(1)
        else:
            yield input_source

    try:
        tts = StreamingTTS()
        
        # Configure TTS based on arguments
        if args.language:
            tts.set_language(args.language)
        if args.voice_gender:
            tts.set_voice_gender(args.voice_gender)
        if args.sample_rate:
            tts.sample_rate = args.sample_rate
        if args.voice_name:
            tts.set_voice(args.voice_name)

        # Process the text
        if args.action == 'speak':
            console.print(
                Panel((
                    f"Saying: {args.text}"
                    ), 
                    title=f"Text to Speech",
                    subtitle=f"{tts.voice_name} is talking"),
                    )
            tts.process_text_stream(
                text_generator(args.text, is_file=args.file)
            )
        elif args.action == 'save':
            if not args.output:
                console.print("Error: Output file path is required for save action")
                return
            
            tts.save_to_file(
                text_generator(args.text, is_file=args.file),
                args.output
            )
            
        console.rule("Successfully processed text-to-speech request.")
            
    except Exception as e:
        console.print(f"[bold red]Error processing text-to-speech: {str(e)}[/bold red]")
        return

def setup_tts_subparser(subparsers):
    """
    Sets up an argparse subparser for the 'tts' command.
    
    Args:
        subparsers: The subparsers object from argparse.ArgumentParser().
    """
    # TTS main parser
    tts_parser = subparsers.add_parser('tts', help='Text-to-Speech conversion utilities')
    tts_subparsers = tts_parser.add_subparsers(dest='action', help='TTS subcommands')

    # Common arguments for both speak and save commands
    common_args = argparse.ArgumentParser(add_help=False)
    common_args.add_argument('text', help='Text to convert to speech (or file path if --file is used)')
    common_args.add_argument('--file', action='store_true', 
                            help='Treat the text argument as a file path')
    common_args.add_argument('--language', default='en-GB',
                            help='Language code (e.g., en-US, es-ES)')
    common_args.add_argument('--voice-gender', choices=['NEUTRAL', 'MALE', 'FEMALE'],
                            default='NEUTRAL', help='Voice gender to use')
    common_args.add_argument('--sample-rate', type=int, default=24000,
                            help='Audio sample rate in Hz')
    common_args.add_argument('--voice_name', default='en-GB-Journey-D', help='A voice name from supported list at https://cloud.google.com/text-to-speech/docs/voices')

    # Speak command - converts text to speech and plays it
    speak_parser = tts_subparsers.add_parser('speak', 
                                            help='Convert text to speech and play it',
                                            parents=[common_args])
    speak_parser.set_defaults(func=tts_command)

    # Save command - converts text to speech and saves to file
    save_parser = tts_subparsers.add_parser('save',
                                           help='Convert text to speech and save to file',
                                           parents=[common_args])
    save_parser.add_argument('--output', default='audio.wav',
                            help='Output audio file path (.wav)')
    save_parser.set_defaults(func=tts_command)

    # Set the default function for the TTS parser
    tts_parser.set_defaults(func=lambda args: tts_parser.print_help())
    




#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
try:
    from langchain_unstructured import UnstructuredLoader
except ImportError:
    UnstructuredLoader = None

from langchain_community.document_loaders import GitLoader
from langchain_community.document_loaders import GoogleDriveLoader

from ..custom_logging import log
from .pdfs import read_pdf_file
from ..utils.config import load_config

import pathlib
import os
import shutil
from urllib.parse import urlparse, unquote
import tempfile
import time

from pydantic import BaseModel, Field
from typing import Optional

UNSTRUCTURED_KEY=os.getenv('UNSTRUCTURED_KEY')

# utility functions
def convert_to_txt(file_path):
    file_dir, file_name = os.path.split(file_path)
    file_base, file_ext = os.path.splitext(file_name)
    txt_file = os.path.join(file_dir, f"{file_base}.txt")
    shutil.copyfile(file_path, txt_file)
    return txt_file


class MyGoogleDriveLoader(GoogleDriveLoader):
    url: Optional[str] = Field(None)

    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, **kwargs, file_ids=['dummy']) # Pass dummy value
        self.url = url

    def _extract_id(self, url):
        parsed_url = urlparse(unquote(url))
        path_parts = parsed_url.path.split('/')
        
        # Iterate over the parts
        for part in path_parts:
            # IDs are typically alphanumeric and at least a few characters long
            # So let's say that to be an ID, a part has to be at least 15 characters long
            if all(char.isalnum() or char in ['_', '-'] for char in part) and len(part) >= 15:
                return part
        
        # Return None if no ID was found
        return None

    def load_from_url(self, url: str):
        id = self._extract_id(url)
        from googleapiclient.errors import HttpError
        from googleapiclient.discovery import build

        # Identify type of URL
        try:
            service = build("drive", "v3", credentials=self._load_credentials())
            file = service.files().get(fileId=id).execute()
        except HttpError as err:
            log.error(f"Error loading file {url}: {str(err)}")
            raise

        mime_type = file["mimeType"]

        if "folder" in mime_type:
            # If it's a folder, load documents from the folder
            return self._load_documents_from_folder(id)
        else:
            # If it's not a folder, treat it as a single file
            if mime_type == "application/vnd.google-apps.document":
                return [self._load_document_from_id(id)]
            elif mime_type == "application/vnd.google-apps.spreadsheet":
                return self._load_sheet_from_id(id)
            elif mime_type == "application/pdf":
                return [self._load_file_from_id(id)]
            else:
                return []

def ignore_files(filepath):
    """Returns True if the given path's file extension is found within 
    config.json "code_extensions" array
    Returns False if not
    """
    # Load the configuration
    config = load_config("config/loader_config.yaml")

    code_extensions = config.get("extensions", [])

    lower_filepath = filepath.lower()
    # TRUE if on the list, FALSE if not
    return any(lower_filepath.endswith(ext) for ext in code_extensions)

def read_git_repo(clone_url, branch="main", metadata=None):
    log.info(f"Reading git repo from {clone_url} - {branch}")
    GIT_PAT = os.getenv('GIT_PAT', None)
    if GIT_PAT is None:
        log.warning("No GIT_PAT is specified, won't be able to clone private git repositories")
    else:
        clone_url = clone_url.replace('https://', f'https://{GIT_PAT}@')
        log.info("Using private GIT_PAT")

    with tempfile.TemporaryDirectory() as tmp_dir:
            try:    
                loader = GitLoader(repo_path=tmp_dir, 
                                   clone_url=clone_url, 
                                   branch=branch,
                                   file_filter=ignore_files)
            except Exception as err:
                log.error(f"Failed to load repository: {str(err)}")
                return None
            docs = loader.load()

            if not docs:
                return None
            
            if metadata is not None:
                for doc in docs:
                    doc.metadata.update(metadata)
            
    log.info(f"GitLoader read {len(docs)} doc(s) from {clone_url}")
        
    return docs


def read_gdrive_to_document(url: str, metadata: dict = None):

    log.info(f"Reading gdrive doc from {url}")

    loader = MyGoogleDriveLoader(url=url)
    docs = loader.load_from_url(url)
    
    if docs is None or len(docs) == 0:
        return None
    
    if metadata is not None:
        for doc in docs:
            doc.metadata.update(metadata)
    
    log.info(f"GoogleDriveLoader read {len(docs)} doc(s) from {url}")

    return docs

def read_url_to_document(url: str, metadata: dict = None):

    if not  UnstructuredLoader:
        raise ImportError("UnstructuredLoader requires 'langchain_unstructured' to be installed")
    unstructured_kwargs = {"skip_infer_table_types": [],
                            "extract_image_block_types":  ["Image", "Table"]
                            }
    loader = UnstructuredLoader(web_url=url, mode="elements", unstructured_kwargs=unstructured_kwargs)
    docs = loader.load()
    if metadata is not None:
        for doc in docs:
            doc.metadata.update(metadata)
            if not doc.metadata.get("source") and doc.metadata.get("url"):
                doc.metadata["source"] = doc.metadata["url"]
    
    log.info(f"UnstructuredLoader docs: {docs}")
    
    return docs

def read_file_to_documents(gs_file: pathlib.Path, metadata: dict = None):
    
    docs = []
    pdf_path = str(pathlib.Path(gs_file))

    # Skip CSV files
    if pdf_path.lower().endswith('.csv'):
        log.info(f"Skipping CSV file: {pdf_path}")
        return []

    if metadata:
        if metadata.get("uploaded_to_bucket"):
            log.info(f"Already uploaded to bucket, skipping {pdf_path}")
            return []

    log.info(f"Sending {pdf_path} to UnstructuredLoader")
    UNSTRUCTURED_URL = os.getenv("UNSTRUCTURED_URL")
    unstructured_kwargs = {"skip_infer_table_types": [],
                            "extract_image_block_types":  ["Image", "Table"]
                            }
    
    if not UnstructuredLoader:
        raise ImportError("UnstructuredLoader requires 'langchain_unstructured' to be installed")
    
    log.info(f"{UNSTRUCTURED_KEY=} and {UNSTRUCTURED_URL=}")
    if UNSTRUCTURED_URL:
        log.info(f"Found UNSTRUCTURED_URL: {UNSTRUCTURED_URL}")
        the_endpoint = f"{UNSTRUCTURED_URL}/general/v0/general"
        try:
            log.info(f"Calling {the_endpoint}")
            loader = UnstructuredLoader(
                file_path=pdf_path,
                url=the_endpoint,
                partition_via_api=True,
                **unstructured_kwargs)
        except Exception as err:
            if "'utf-8' codec can't decode byte" in str(err):
                log.error(f"Unstructured can't decode byte for {gs_file}: {str(err)} - aborting")
                return []
            else:
                raise err
    else:
        log.info("Calling Unstructured Loader via Public API")
        loader = UnstructuredLoader(
            file_path=pdf_path,
            api_key=UNSTRUCTURED_KEY,
            mode="elements",
            **unstructured_kwargs)
    
    start = time.time()
    try:
        docs = loader.load() # this takes a long time 30m+ for big PDF files
    except ValueError as e:
        log.info(f"Error for {gs_file} from UnstructuredLoader: {str(e)}")
        pdf_path = pathlib.Path(gs_file)
        if pdf_path.suffix == ".pdf":
            local_doc = read_pdf_file(pdf_path, metadata=metadata)
            if local_doc is not None:
                docs.append(local_doc)

        elif "file type is not supported in partition" in str(e):
            txt_docs = convert_to_txt_and_extract(gs_file, split=False)
            if txt_docs:
                docs.extend(txt_docs)
            else:
                log.warning("Could not extract any docs for {gs_file}")
    
    end = time.time()
    elapsed_time = round((end - start) / 60, 2)
    log.info(f"Loaded docs for {gs_file} from read_file_to_docs took {elapsed_time} mins")

    if not isinstance(docs, list):
        log.warning(f"read_file_to_documents for {gs_file} returned not a list got {type(docs)})")
        docs = [docs] 

    total_doc_chars = 0
    doc_chunk = 0
    for doc in docs:
        log.info(f"{gs_file} doc_content: {doc.page_content[:30]} - length: {len(doc.page_content)}")
        total_doc_chars += len(doc.page_content)
        doc.metadata["doc_chunk"] = doc_chunk
        doc_chunk += 1
        if metadata is not None:
            doc.metadata.update(metadata)
            log_meta = metadata.copy()
            if 'image_base64' in log_meta:
                log_meta['image_base64'] = "XXXtruncatedXXXX"
            log.info(f"doc_metadata: {log_meta}")
    
    for doc in docs:
        doc.metadata["parse_total_doc_chars"] = total_doc_chars
        doc.metadata["parse_total_doc_chunks"] = doc_chunk
    
    log.info(f"gs_file:{gs_file} read into {len(docs)} docs. parse_total_doc_chars: {total_doc_chars} - parse_total_doc_chunks {doc_chunk}")

    return docs

def convert_to_txt_and_extract(gs_file, split=False):

    if not  UnstructuredLoader:
        raise ImportError("UnstructuredLoader requires 'langchain_unstructured' to be installed")
    
    log.info("trying file parsing locally via .txt conversion")
    txt_file = None
    docs = []
    try:
        # Convert the file to .txt and try again
        txt_file = convert_to_txt(gs_file)
        loader = UnstructuredLoader(
            txt_file, 
            mode="elements")

        docs = loader.load_and_split() if split else loader.load()

    except Exception as inner_e:
        raise Exception("An error occurred during txt conversion or loading.") from inner_e

    finally:
        # Ensure cleanup happens if txt_file was created
        if txt_file is not None and os.path.exists(txt_file):
            os.remove(txt_file)
    
    return docs